@@ -6,7 +6,7 @@
 #include <unistd.h>
 #include <errno.h>
 #include <sys/sysctl.h>
-#if HAVE_SYS_VMMETER_H
+#ifdef HAVE_SYS_VMMETER_H
 #include <sys/vmmeter.h>
 #endif
 #include <sys/swap.h>
@@ -35,39 +35,60 @@ int netsnmp_mem_arch_load( netsnmp_cache *cache, void *magic ) {
 
     struct uvmexp  uvmexp;
     size_t         uvmexp_size  = sizeof(uvmexp);
-    int            uvmexp_mib[] = { CTL_VM, VM_UVMEXP };
 
     struct vmtotal total;
     size_t         total_size  = sizeof(total);
-    int            total_mib[] = { CTL_VM, VM_METER };
 
     quad_t          phys_mem;
     quad_t          user_mem;
-    unsigned int    bufspace;
-    unsigned int    maxbufspace;
     size_t          mem_size  = sizeof(phys_mem);
+#if defined(openbsd)
+    int             phys_mem_mib[] = { CTL_HW, HW_PHYSMEM64 };
+    int             user_mem_mib[] = { CTL_HW, HW_USERMEM64 };
+    int             uvmexp_mib[] = { CTL_VM, VM_UVMEXP };
+    int             total_mib[] = { CTL_VM, VM_METER };
+#else
+    unsigned long   bufspace;
+    unsigned long   maxbufspace;
     size_t          buf_size  = sizeof(bufspace);
-    int             phys_mem_mib[] = { CTL_HW, HW_PHYSMEM };
-    int             user_mem_mib[] = { CTL_HW, HW_USERMEM };
+#endif
 
     /*
      * Retrieve the memory information from the underlying O/S...
      */
-#if defined(__NetBSD__)
-    sysctlbyname("vm.uvmexp",              &uvmexp, &uvmexp_size, NULL, 0);
-    sysctlbyname("vm.vmmeter",              &total,  &total_size, NULL, 0);
-    sysctlbyname("hw.physmem64",         &phys_mem,    &mem_size, NULL, 0);
-    sysctlbyname("hw.usermem64",         &user_mem,    &mem_size, NULL, 0);
-    sysctlbyname("vm.bufmem",            &bufspace,    &buf_size, NULL, 0);
-    sysctlbyname("vm.bufmem_hiwater", &maxbufspace,    &buf_size, NULL, 0);
-    pagesize = sysconf(_SC_PAGESIZE);
-#else  /* do we still need this section ? */
+#if defined(openbsd)
     sysctl(uvmexp_mib,   2, &uvmexp,   &uvmexp_size,   NULL, 0);
     sysctl(total_mib,    2, &total,    &total_size,    NULL, 0);
     sysctl(phys_mem_mib, 2, &phys_mem, &mem_size,      NULL, 0);
     sysctl(user_mem_mib, 2, &user_mem, &mem_size,      NULL, 0);
-    pagesize = uvmexp.pagesize;
+#else
+    if (sysctlbyname("vm.uvmexp", &uvmexp, &uvmexp_size, NULL, 0) == -1) {
+        snmp_log_perror("sysctl vm.uvmexp");
+        return -1;
+    }
+    if (sysctlbyname("vm.vmmeter", &total,  &total_size, NULL, 0) == -1) {
+        snmp_log_perror("sysctl vm.vmmeter");
+        return -1;
+    }
+    if (sysctlbyname("hw.physmem64", &phys_mem, &mem_size, NULL, 0) == -1) {
+        snmp_log_perror("sysctl hw.physmem64");
+        return -1;
+    }
+    if (sysctlbyname("hw.usermem64", &user_mem, &mem_size, NULL, 0) == -1) {
+        snmp_log_perror("sysctl hw.usermem64");
+        return -1;
+    }
+    if (sysctlbyname("vm.bufmem", &bufspace, &buf_size, NULL, 0) == -1) {
+        snmp_log_perror("sysctl vm.bufmem");
+        return -1;
+    }
+    if (sysctlbyname("vm.bufmem_hiwater", &maxbufspace, &buf_size, NULL, 0) == -1) {
+        snmp_log_perror("sysctl vm.bufmem_hiwater");
+        return -1;
+    }
+
 #endif
+    pagesize = sysconf(_SC_PAGESIZE);
 
     /*
      * ... and save this in a standard form.
@@ -94,7 +115,7 @@ int netsnmp_mem_arch_load( netsnmp_cache *cache, void *magic ) {
         mem->free  = uvmexp.free;
     }
 
-#if defined(__NetBSD__)
+#if 1
     mem = netsnmp_memory_get_byIdx( NETSNMP_MEM_TYPE_VIRTMEM, 1 );
     if (!mem) {
         snmp_log_perror("No Virtual Memory info entry");
@@ -190,7 +211,7 @@ swapinfo(long pagesize)
             continue;
         if (!mem->descr) {
          /* sprintf(buf, "swap #%d", s[i].se_dev); */
-            sprintf(buf, "swap %s",  s[i].se_path);
+            snprintf(buf, sizeof buf, "swap %s",  s[i].se_path);
             mem->descr = strdup( buf );
         }
         mem->units = pagesize;
