@@ -1,26 +1,30 @@
 #include <net-snmp/net-snmp-config.h>
+#include <net-snmp/net-snmp-features.h>
 
-#if HAVE_STDLIB_H
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
-#if HAVE_UNISTD_H
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 #include <stdio.h>
-#if HAVE_STRING_H
+#ifdef HAVE_STRING_H
 #include <string.h>
 #else
 #include <strings.h>
 #endif
 #include <ctype.h>
 #include <sys/types.h>
-#if HAVE_WINSOCK_H
-#include <winsock.h>
-#else
+#ifdef HAVE_ARPA_INET_H
+#include <arpa/inet.h>
+#endif
+#ifdef HAVE_NETINET_IN_H
 #include <netinet/in.h>
+#endif
+#ifdef HAVE_NETDB_H
 #include <netdb.h>
 #endif
-#if HAVE_SYS_WAIT_H
+#ifdef HAVE_SYS_WAIT_H
 #include <sys/wait.h>
 #endif
 
@@ -37,10 +41,17 @@
 #include "snmptrapd_log.h"
 #include "notification-log-mib/notification_log.h"
 
+netsnmp_feature_child_of(add_default_traphandler, snmptrapd);
+
 char *syslog_format1 = NULL;
 char *syslog_format2 = NULL;
 char *print_format1  = NULL;
 char *print_format2  = NULL;
+char *exec_format1   = NULL;
+char *exec_format2   = NULL;
+
+int   SyslogTrap = 0;
+int   dropauth = 0;
 
 const char     *trap1_std_str = "%.4y-%.2m-%.2l %.2h:%.2j:%.2k %B [%b] (via %A [%a]): %N\n\t%W Trap (%q) Uptime: %#T\n%v\n";
 const char     *trap2_std_str = "%.4y-%.2m-%.2l %.2h:%.2j:%.2k %B [%b]:\n%v\n";
@@ -81,19 +92,28 @@ snmptrapd_parse_traphandle(const char *token, char *line)
     char           *cptr, *cp;
     netsnmp_trapd_handler *traph;
     int             flags = 0;
+    char           *format = NULL;
 
     memset( buf, 0, sizeof(buf));
     memset(obuf, 0, sizeof(obuf));
     cptr = copy_nword(line, buf, sizeof(buf));
+
+    if ( buf[0] == '-' && buf[1] == 'F' ) {
+        cptr = copy_nword(cptr, buf, sizeof(buf));
+        format = strdup( buf );
+        cptr = copy_nword(cptr, buf, sizeof(buf));
+    }
     if ( !cptr ) {
-        config_perror("Missing traphandle command");
+        netsnmp_config_error("Missing traphandle command (%s)", buf);
+        free(format);
         return;
     }
 
     DEBUGMSGTL(("read_config:traphandle", "registering handler for: "));
     if (!strcmp(buf, "default")) {
         DEBUGMSG(("read_config:traphandle", "default"));
-        traph = netsnmp_add_default_traphandler( command_handler );
+        traph = netsnmp_add_global_traphandler(NETSNMPTRAPD_DEFAULT_HANDLER,
+                                               command_handler );
     } else {
         cp = buf+strlen(buf)-1;
         if ( *cp == '*' ) {
@@ -108,11 +128,9 @@ snmptrapd_parse_traphandle(const char *token, char *line)
             }
         }
         if (!read_objid(buf, obuf, &olen)) {
-            char            buf1[STRINGMAX];
-            snprintf(buf1,  sizeof(buf1),
-                    "Bad trap OID in traphandle directive: %s", buf);
-            buf1[ sizeof(buf1)-1 ] = 0;
-            config_perror(buf1);
+	    netsnmp_config_error("Bad trap OID in traphandle directive: %s",
+				 buf);
+            free(format);
             return;
         }
         DEBUGMSGOID(("read_config:traphandle", obuf, olen));
@@ -125,7 +143,12 @@ snmptrapd_parse_traphandle(const char *token, char *line)
         traph->flags = flags;
         traph->authtypes = TRAP_AUTH_EXE;
         traph->token = strdup(cptr);
+        if (format) {
+            traph->format = format;
+            format = NULL;
+        }
     }
+    free(format);
 }
 
 
@@ -138,17 +161,26 @@ parse_forward(const char *token, char *line)
     char           *cptr, *cp;
     netsnmp_trapd_handler *traph;
     int             flags = 0;
+    char           *format = NULL;
 
     memset( buf, 0, sizeof(buf));
     memset(obuf, 0, sizeof(obuf));
     cptr = copy_nword(line, buf, sizeof(buf));
+
+    if ( buf[0] == '-' && buf[1] == 'F' ) {
+        cptr = copy_nword(cptr, buf, sizeof(buf));
+        format = strdup( buf );
+        cptr = copy_nword(cptr, buf, sizeof(buf));
+    }
     DEBUGMSGTL(("read_config:forward", "registering forward for: "));
     if (!strcmp(buf, "default")) {
         DEBUGMSG(("read_config:forward", "default"));
         if ( !strcmp( cptr, "agentx" ))
-            traph = netsnmp_add_default_traphandler( axforward_handler );
+            traph = netsnmp_add_global_traphandler(NETSNMPTRAPD_DEFAULT_HANDLER,
+                                            axforward_handler );
         else
-            traph = netsnmp_add_default_traphandler( forward_handler );
+            traph = netsnmp_add_global_traphandler(NETSNMPTRAPD_DEFAULT_HANDLER,
+                                            forward_handler );
     } else {
         cp = buf+strlen(buf)-1;
         if ( *cp == '*' ) {
@@ -164,11 +196,8 @@ parse_forward(const char *token, char *line)
         }
 
         if (!read_objid(buf, obuf, &olen)) {
-            char            buf1[STRINGMAX];
-            snprintf(buf1,  sizeof(buf1),
-                    "Bad trap OID in forward directive: %s", buf);
-            buf1[ sizeof(buf1)-1 ] = 0;
-            config_perror(buf1);
+	    netsnmp_config_error("Bad trap OID in forward directive: %s", buf);
+            free(format);
             return;
         }
         DEBUGMSGOID(("read_config:forward", obuf, olen));
@@ -184,21 +213,25 @@ parse_forward(const char *token, char *line)
         traph->flags = flags;
         traph->authtypes = TRAP_AUTH_NET;
         traph->token = strdup(cptr);
+        if (format)
+            traph->format = format;
+    } else {
+        free(format);
     }
 }
 
 
-static void
+void
 parse_format(const char *token, char *line)
 {
-    char *cp;
+    char *cp, *sep;
 
     /*
      * Extract the first token from the value
      * which tells us which style of format this is
      */
     cp = line;
-    while (*cp && !isspace(*cp))
+    while (*cp && !isspace((unsigned char)(*cp)))
         cp++;
     if (!(*cp)) {
         /*
@@ -209,8 +242,9 @@ parse_format(const char *token, char *line)
         return;
     }
 
+    sep = cp;
     *(cp++) = '\0';
-    while (*cp && !isspace(*cp))
+    while (*cp && isspace((unsigned char)(*cp)))
         cp++;
 
     /*
@@ -240,7 +274,20 @@ parse_format(const char *token, char *line)
         SNMP_FREE( syslog_format2 );
         syslog_format1 = strdup(cp);
         syslog_format2 = strdup(cp);
+    } else if (!strcmp( line, "execute1")) {
+        SNMP_FREE( exec_format1 );
+        exec_format1 = strdup(cp);
+    } else if (!strcmp( line, "execute2")) {
+        SNMP_FREE( exec_format2 );
+        exec_format2 = strdup(cp);
+    } else if (!strcmp( line, "execute")) {
+        SNMP_FREE( exec_format1 );
+        SNMP_FREE( exec_format2 );
+        exec_format1 = strdup(cp);
+        exec_format2 = strdup(cp);
     }
+
+    *sep = ' ';
 }
 
 
@@ -255,7 +302,7 @@ void
 free_trap1_fmt(void)
 {
     if (print_format1 && print_format1 != trap1_std_str)
-        free((char *) print_format1);
+        free(print_format1);
     print_format1 = NULL;
 }
 
@@ -271,7 +318,7 @@ void
 free_trap2_fmt(void)
 {
     if (print_format2 && print_format2 != trap2_std_str)
-        free((char *) print_format2);
+        free(print_format2);
     print_format2 = NULL;
 }
 
@@ -289,7 +336,7 @@ snmptrapd_register_configs( void )
                             parse_trap2_fmt, free_trap2_fmt, "format");
     register_config_handler("snmptrapd", "format",
                             parse_format, NULL,
-			    "[print{,1,2}|syslog{,1,2}] format");
+			    "[print{,1,2}|syslog{,1,2}|execute{,1,2}] format");
     register_config_handler("snmptrapd", "forward",
                             parse_forward, NULL, "OID|\"default\" destination");
 }
@@ -308,12 +355,26 @@ netsnmp_trapd_handler *netsnmp_post_global_traphandlers   = NULL;
 netsnmp_trapd_handler *netsnmp_default_traphandlers  = NULL;
 netsnmp_trapd_handler *netsnmp_specific_traphandlers = NULL;
 
+typedef struct netsnmp_handler_map_t {
+   netsnmp_trapd_handler **handler;
+   const char             *descr;
+} netsnmp_handler_map;
+
+static netsnmp_handler_map handlers[] = {
+    { &netsnmp_auth_global_traphandlers, "auth trap" },
+    { &netsnmp_pre_global_traphandlers, "pre-global trap" },
+    { NULL, "trap specific" },
+    { &netsnmp_post_global_traphandlers, "global" },
+    { NULL, NULL }
+};
+
 /*
  * Register a new "global" traphandler,
  * to be applied to *all* incoming traps
  */
 netsnmp_trapd_handler *
-netsnmp_add_global_traphandler(int list, Netsnmp_Trap_Handler* handler) {
+netsnmp_add_global_traphandler(int list, Netsnmp_Trap_Handler *handler)
+{
     netsnmp_trapd_handler *traph;
 
     if ( !handler )
@@ -324,7 +385,7 @@ netsnmp_add_global_traphandler(int list, Netsnmp_Trap_Handler* handler) {
         return NULL;
 
     /*
-     * Add this new handler to the front of the appropriate global list
+     * Add this new handler to the front of the appropriate list
      *   (or should it go on the end?)
      */
     traph->handler = handler;
@@ -342,6 +403,10 @@ netsnmp_add_global_traphandler(int list, Netsnmp_Trap_Handler* handler) {
         traph->nexth   = netsnmp_post_global_traphandlers;
         netsnmp_post_global_traphandlers = traph;
         break;
+    case NETSNMPTRAPD_DEFAULT_HANDLER:
+        traph->nexth   = netsnmp_default_traphandlers;
+        netsnmp_default_traphandlers = traph;
+        break;
     default:
         free( traph );
         return NULL;
@@ -349,32 +414,17 @@ netsnmp_add_global_traphandler(int list, Netsnmp_Trap_Handler* handler) {
     return traph;
 }
 
-
+#ifndef NETSNMP_FEATURE_REMOVE_ADD_DEFAULT_TRAPHANDLER
 /*
  * Register a new "default" traphandler, to be applied to all
  * traps with no specific trap handlers of their own.
  */
 netsnmp_trapd_handler *
-netsnmp_add_default_traphandler( Netsnmp_Trap_Handler* handler) {
-    netsnmp_trapd_handler *traph;
-
-    if ( !handler )
-        return NULL;
-
-    traph = SNMP_MALLOC_TYPEDEF(netsnmp_trapd_handler);
-    if ( !traph )
-        return NULL;
-
-    /*
-     * Add this new handler to the front of the default list
-     *   (or should it go on the end?)
-     */
-    traph->authtypes = TRAP_AUTH_ALL; /* callers will likely change this */
-    traph->handler = handler;
-    traph->nexth   = netsnmp_default_traphandlers;
-    netsnmp_default_traphandlers = traph;
-    return traph;
+netsnmp_add_default_traphandler(Netsnmp_Trap_Handler *handler) {
+    return netsnmp_add_global_traphandler(NETSNMPTRAPD_DEFAULT_HANDLER,
+                                          handler);
 }
+#endif /* NETSNMP_FEATURE_REMOVE_ADD_DEFAULT_TRAPHANDLER */
 
 
 /*
@@ -399,8 +449,7 @@ netsnmp_add_traphandler(Netsnmp_Trap_Handler* handler,
     traph->authtypes   = TRAP_AUTH_ALL; /* callers will likely change this */
     traph->handler     = handler;
     traph->trapoid_len = trapOidLen;
-    memdup((u_char **)&(traph->trapoid), (u_char *)trapOid,
-		    sizeof(oid) * trapOidLen);
+    traph->trapoid     = snmp_duplicate_objid(trapOid, trapOidLen);
 
     /*
      * Now try to find the appropriate place in the trap-specific
@@ -515,8 +564,13 @@ netsnmp_trapd_handler *
 netsnmp_get_traphandler( oid *trapOid, int trapOidLen ) {
     netsnmp_trapd_handler *traph;
     
-    if (!trapOid || !trapOidLen)
+    if (!trapOid || !trapOidLen) {
+        DEBUGMSGTL(( "snmptrapd:lookup", "get_traphandler no OID!\n"));
         return NULL;
+    }
+    DEBUGMSGTL(( "snmptrapd:lookup", "Looking up Trap OID: "));
+    DEBUGMSGOID(("snmptrapd:lookup", trapOid, trapOidLen));
+    DEBUGMSG(( "snmptrapd:lookup", "\n"));
 
     /*
      * Look for a matching OID, and return that list...
@@ -531,7 +585,8 @@ netsnmp_get_traphandler( oid *trapOid, int trapOidLen ) {
         if (!(traph->flags & NETSNMP_TRAPHANDLER_FLAG_MATCH_TREE)) {
             if (snmp_oid_compare(traph->trapoid, traph->trapoid_len,
                                  trapOid, trapOidLen) == 0) {
-                DEBUGMSGTL(( "snmptrapd", "get_traphandler exact match (%x)\n", traph));
+                DEBUGMSGTL(( "snmptrapd:lookup",
+                             "get_traphandler exact match (%p)\n", traph));
 	        return traph;
             }
 	} else {
@@ -549,11 +604,11 @@ netsnmp_get_traphandler( oid *trapOid, int trapOidLen ) {
                      */
                     if (snmp_oid_compare(traph->trapoid, traph->trapoid_len,
                                          trapOid, trapOidLen) != 0) {
-                        DEBUGMSGTL(( "snmptrapd", "get_traphandler strict subtree match (%x)\n", traph));
+                        DEBUGMSGTL(( "snmptrapd:lookup", "get_traphandler strict subtree match (%p)\n", traph));
 	                return traph;
                     }
                 } else {
-                    DEBUGMSGTL(( "snmptrapd", "get_traphandler subtree match (%x)\n", traph));
+                    DEBUGMSGTL(( "snmptrapd:lookup", "get_traphandler subtree match (%p)\n", traph));
 	            return traph;
                 }
             }
@@ -563,7 +618,7 @@ netsnmp_get_traphandler( oid *trapOid, int trapOidLen ) {
     /*
      * .... or failing that, return the "default" list (which may be NULL)
      */
-    DEBUGMSGTL(( "snmptrapd", "get_traphandler default (%x)\n",
+    DEBUGMSGTL(( "snmptrapd:lookup", "get_traphandler default (%p)\n",
 			    netsnmp_default_traphandlers));
     return netsnmp_default_traphandlers;
 }
@@ -588,7 +643,6 @@ int   syslog_handler(  netsnmp_pdu           *pdu,
     u_char         *rbuf = NULL;
     size_t          r_len = 64, o_len = 0;
     int             trunc = 0;
-    extern int      SyslogTrap;
 
     DEBUGMSGTL(( "snmptrapd", "syslog_handler\n"));
 
@@ -671,7 +725,6 @@ int   print_handler(   netsnmp_pdu           *pdu,
     u_char         *rbuf = NULL;
     size_t          r_len = 64, o_len = 0;
     int             trunc = 0;
-    extern int      dropauth;
 
     DEBUGMSGTL(( "snmptrapd", "print_handler\n"));
 
@@ -739,163 +792,20 @@ int   print_handler(   netsnmp_pdu           *pdu,
 }
 
 
+#define EXECUTE_FORMAT	"%B\n%b\n%V\n%v\n"
+
 /*
  *  Trap handler for invoking a suitable script
  */
-void
-send_handler_data(FILE * file, struct hostent *host,
-                  netsnmp_pdu *pdu, netsnmp_transport *transport)
-{
-    netsnmp_variable_list tmpvar, *vars;
-    static oid      trapoids[] = { 1, 3, 6, 1, 6, 3, 1, 1, 5, 0 };
-    static oid      snmpsysuptime[] = { 1, 3, 6, 1, 2, 1, 1, 3, 0 };
-    static oid      snmptrapoid[] = { 1, 3, 6, 1, 6, 3, 1, 1, 4, 1, 0 };
-    static oid      snmptrapent[] = { 1, 3, 6, 1, 6, 3, 1, 1, 4, 3, 0 };
-    static oid      snmptrapaddr[] = { 1, 3, 6, 1, 6, 3, 18, 1, 3, 0 };
-    static oid      snmptrapcom[] = { 1, 3, 6, 1, 6, 3, 18, 1, 4, 0 };
-    oid             enttrapoid[MAX_OID_LEN];
-    int             enttraplen = pdu->enterprise_length;
-
-    if (transport != NULL && transport->f_fmtaddr != NULL) {
-        char *tstr = transport->f_fmtaddr(transport, pdu->transport_data,
-                                          pdu->transport_data_length);
-        if (tstr != NULL) {
-            fprintf(file, "%s\n%s\n", host ? host->h_name : tstr, tstr);
-            free(tstr);
-        }
-    } else {
-        fprintf(file, "%s\n<UNKNOWN>\n", host ? host->h_name : "<UNKNOWN>");
-    }
-    if (pdu->command == SNMP_MSG_TRAP) {
-        /*
-         * convert a v1 trap to a v2 variable binding list:
-         * The uptime and trapOID go first in the list. 
-         */
-        tmpvar.val.integer = (long *) &pdu->time;
-        tmpvar.val_len = sizeof(pdu->time);
-        tmpvar.type = ASN_TIMETICKS;
-        fprint_variable(file, snmpsysuptime,
-                        sizeof(snmpsysuptime) / sizeof(oid), &tmpvar);
-        tmpvar.type = ASN_OBJECT_ID;
-        if (pdu->trap_type == SNMP_TRAP_ENTERPRISESPECIFIC) {
-            memcpy(enttrapoid, pdu->enterprise, sizeof(oid) * enttraplen);
-            if (enttrapoid[enttraplen - 1] != 0)
-                enttrapoid[enttraplen++] = 0;
-            enttrapoid[enttraplen++] = pdu->specific_type;
-            tmpvar.val.objid = enttrapoid;
-            tmpvar.val_len = enttraplen * sizeof(oid);
-        } else {
-            trapoids[9] = pdu->trap_type + 1;
-            tmpvar.val.objid = trapoids;
-            tmpvar.val_len = 10 * sizeof(oid);
-        }
-        fprint_variable(file, snmptrapoid,
-                        sizeof(snmptrapoid) / sizeof(oid), &tmpvar);
-    }
-    /*
-     * do the variables in the pdu 
-     */
-    for (vars = pdu->variables; vars; vars = vars->next_variable) {
-        fprint_variable(file, vars->name, vars->name_length, vars);
-    }
-    if (pdu->command == SNMP_MSG_TRAP) {
-        /*
-         * convert a v1 trap to a v2 variable binding list:
-         * The enterprise goes last. 
-         */
-        tmpvar.val.string = pdu->agent_addr;
-        tmpvar.val_len = 4;
-        tmpvar.type = ASN_IPADDRESS;
-        fprint_variable(file, snmptrapaddr,
-                        sizeof(snmptrapaddr) / sizeof(oid), &tmpvar);
-        tmpvar.val.string = pdu->community;
-        tmpvar.val_len = pdu->community_len;
-        tmpvar.type = ASN_OCTET_STR;
-        fprint_variable(file, snmptrapcom,
-                        sizeof(snmptrapcom) / sizeof(oid), &tmpvar);
-        tmpvar.val.objid = pdu->enterprise;
-        tmpvar.val_len = pdu->enterprise_length * sizeof(oid);
-        tmpvar.type = ASN_OBJECT_ID;
-        fprint_variable(file, snmptrapent,
-                        sizeof(snmptrapent) / sizeof(oid), &tmpvar);
-    }
-}
-
-void
-do_external(char *cmd, struct hostent *host,
-            netsnmp_pdu *pdu, netsnmp_transport *transport)
-{
-    FILE           *file;
-    int             oldquick, result;
-
-    DEBUGMSGTL(("snmptrapd", "Running: %s\n", cmd));
-    oldquick = snmp_get_quick_print();
-    snmp_set_quick_print(1);
-    if (cmd) {
-#ifndef WIN32
-        int             fd[2];
-        int             pid;
-
-        if (pipe(fd)) {
-            snmp_log_perror("pipe");
-        }
-        if ((pid = fork()) == 0) {
-            /*
-             * child 
-             */
-            close(0);
-            if (dup(fd[0]) != 0) {
-                snmp_log_perror("dup");
-            }
-            close(fd[1]);
-            close(fd[0]);
-            system(cmd);
-            exit(0);
-        } else if (pid > 0) {
-            file = fdopen(fd[1], "w");
-            send_handler_data(file, host, pdu, transport);
-            fclose(file);
-            close(fd[0]);
-            close(fd[1]);
-            if (waitpid(pid, &result, 0) < 0) {
-                snmp_log_perror("waitpid");
-            }
-        } else {
-            snmp_log_perror("fork");
-        }
-#else
-        char            command_buf[128];
-        char            file_buf[L_tmpnam];
-
-        tmpnam(file_buf);
-        file = fopen(file_buf, "w");
-        if (!file) {
-            fprintf(stderr, "fopen: %s: %s\n", file_buf, strerror(errno));
-        } else {
-            send_handler_data(file, host, pdu, transport);
-            fclose(file);
-            snprintf(command_buf, sizeof(command_buf),
-                     "%s < %s", cmd, file_buf);
-            command_buf[ sizeof(command_buf)-1 ] = 0;
-            result = system(command_buf);
-            if (result == -1)
-                fprintf(stderr, "system: %s: %s\n", command_buf,
-                        strerror(errno));
-            else if (result)
-                fprintf(stderr, "system: %s: %d\n", command_buf, result);
-            remove(file_buf);
-        }
-#endif                          /* WIN32 */
-    }
-    snmp_set_quick_print(oldquick);
-}
-
-#define EXECUTE_FORMAT	"%B\n%b\n%V\n%v\n"
-
 int   command_handler( netsnmp_pdu           *pdu,
                        netsnmp_transport     *transport,
                        netsnmp_trapd_handler *handler)
 {
+#ifndef USING_UTILITIES_EXECUTE_MODULE
+    NETSNMP_LOGONCE((LOG_WARNING,
+                     "support for run_shell_command not available\n"));
+    return NETSNMPTRAPD_HANDLER_FAIL;
+#else
     u_char         *rbuf = NULL;
     size_t          r_len = 64, o_len = 0;
     int             oldquick;
@@ -908,8 +818,10 @@ int   command_handler( netsnmp_pdu           *pdu,
 	    v2_pdu = convert_v1pdu_to_v2(pdu);
 	else
 	    v2_pdu = pdu;
-        oldquick = snmp_get_quick_print();
-        snmp_set_quick_print(1);
+        oldquick = netsnmp_ds_get_boolean(NETSNMP_DS_LIBRARY_ID, 
+                                          NETSNMP_DS_LIB_QUICK_PRINT);
+        netsnmp_ds_set_boolean(NETSNMP_DS_LIBRARY_ID, 
+                               NETSNMP_DS_LIB_QUICK_PRINT, 1);
 
         /*
 	 * Format the trap and pass this string to the external command
@@ -923,50 +835,44 @@ int   command_handler( netsnmp_pdu           *pdu,
          *  If there's a format string registered for this trap, then use it.
          *  Otherwise use the standard execution format setting.
          */
-        if (handler && handler->format && *handler->format) {
+        if (handler->format && *handler->format) {
             DEBUGMSGTL(( "snmptrapd", "format = '%s'\n", handler->format));
             realloc_format_trap(&rbuf, &r_len, &o_len, 1,
                                              handler->format,
                                              v2_pdu, transport);
         } else {
-            DEBUGMSGTL(( "snmptrapd", "execute format\n"));
-            realloc_format_trap(&rbuf, &r_len, &o_len, 1,
-                                             EXECUTE_FORMAT,
+	    if ( pdu->command == SNMP_MSG_TRAP && exec_format1 ) {
+                DEBUGMSGTL(( "snmptrapd", "exec v1 = '%s'\n", exec_format1));
+                realloc_format_trap(&rbuf, &r_len, &o_len, 1,
+                                             exec_format1, pdu, transport);
+	    } else if ( pdu->command != SNMP_MSG_TRAP && exec_format2 ) {
+                DEBUGMSGTL(( "snmptrapd", "exec v2/3 = '%s'\n", exec_format2));
+                realloc_format_trap(&rbuf, &r_len, &o_len, 1,
+                                             exec_format2, pdu, transport);
+	    } else {
+                DEBUGMSGTL(( "snmptrapd", "execute format\n"));
+                realloc_format_trap(&rbuf, &r_len, &o_len, 1, EXECUTE_FORMAT,
                                              v2_pdu, transport);
+            }
 	}
 
         /*
          *  and pass this formatted string to the command specified
          */
-        run_shell_command(handler->token, (char*)rbuf, NULL, 0);   /* Not interested in output */
-        snmp_set_quick_print(oldquick);
+        run_shell_command(handler->token, (char*)rbuf, NULL, NULL);   /* Not interested in output */
+        netsnmp_ds_set_boolean(NETSNMP_DS_LIBRARY_ID, 
+                               NETSNMP_DS_LIB_QUICK_PRINT, oldquick);
         if (pdu->command == SNMP_MSG_TRAP)
             snmp_free_pdu(v2_pdu);
         free(rbuf);
     }
     return NETSNMPTRAPD_HANDLER_OK;
+#endif /* !def USING_UTILITIES_EXECUTE_MODULE */
 }
 
 
 
 
-/*
- *  Trap handler for doing something with "event" traps
- *      (not entirely clear what this is about ???)
- */
-	/* XXX - in snmptrapd.c */
-void event_input(netsnmp_variable_list * vp);
-
-int   event_handler( netsnmp_pdu           *pdu,
-                     netsnmp_transport     *transport,
-                     netsnmp_trapd_handler *handler)
-{
-    DEBUGMSGTL(( "snmptrapd", "event_handler\n"));
-    event_input(pdu->variables);
-    return NETSNMPTRAPD_HANDLER_OK;
-}
-
-
 /*
  *  Trap handler for forwarding to the AgentX master agent
  */
@@ -978,6 +884,121 @@ int axforward_handler( netsnmp_pdu           *pdu,
     return NETSNMPTRAPD_HANDLER_OK;
 }
 
+static int add_forwarder_info(netsnmp_pdu *pdu, netsnmp_pdu *pdu2)
+{
+    netsnmp_indexed_addr_pair *addr_pair1;
+    struct sockaddr_in *to1 = NULL;
+    struct sockaddr_in *to2 = NULL;
+    int            last_snmpTrapAddress_index = -1;
+    /* snmpTrapAddress_oid.0 is also defined as agentaddr_oid */
+    const oid      snmpTrapAddress_oid[]   = { 1,3,6,1,6,3,18,1,3};
+    /* each forwarder will add this OID with changed last index */
+    oid            forwarder_oid[]   =       { 1,3,6,1,6,3,18,1,3,0};
+    const size_t   snmpTrapAddress_oid_size = OID_LENGTH(snmpTrapAddress_oid);
+    const size_t   forwarder_oid_len = OID_LENGTH(forwarder_oid);
+    struct in_addr agent_addr;
+    struct in_addr my_ip_addr;
+
+    memset(&agent_addr, 0, sizeof(agent_addr));
+    memset(&my_ip_addr, 0, sizeof(my_ip_addr));
+
+    if (pdu && pdu->transport_data &&
+        pdu->transport_data_length == sizeof(*addr_pair1)) {
+        addr_pair1 = (netsnmp_indexed_addr_pair *)pdu->transport_data;
+
+        /*
+         * Get the IPv4 address of the host that this trap was sent from =
+         * last forwarder's IP address.
+         */
+        if (addr_pair1->remote_addr.sa.sa_family == AF_INET) {
+            to1 = (struct sockaddr_in *)&(addr_pair1->remote_addr);
+            agent_addr = to1->sin_addr;
+        }
+        /*
+         * Get the IPv4 address of the host that this trap was sent to =
+         * this forwarder's IP address.
+         */
+        if (addr_pair1->local_addr.sa.sa_family == AF_INET) {
+            to2 = (struct sockaddr_in *)&(addr_pair1->local_addr);
+            my_ip_addr = to2->sin_addr;
+        }
+    }
+
+    if (to1) {
+        netsnmp_variable_list *vblist = NULL;
+        netsnmp_variable_list *var = NULL;
+
+        if (*(in_addr_t *)pdu2->agent_addr == INADDR_ANY) {
+            /*
+             * there was no agent address defined in PDU. copy the forwarding
+             * agent IP address from the transport socket.
+             */
+            *(struct in_addr *)pdu2->agent_addr = agent_addr;
+        }
+
+        vblist = pdu2->variables;
+
+        /*
+         * Iterate over all varbinds in the PDU to see if it already has any
+         * forwarder information.
+         */
+        for (var = vblist; var; var = var->next_variable) {
+            if (snmp_oid_ncompare(var->name, var->name_length,
+                                  snmpTrapAddress_oid,
+                                  snmpTrapAddress_oid_size,
+                                  snmpTrapAddress_oid_size) == 0) {
+                int my_snmpTrapAddress_index =
+                    var->name[var->name_length - 1];
+
+                DEBUGMSGTL(("snmptrapd", "  my_snmpTrapAddress_index=%d, last_snmpTrapAddress_index=%d, my_ip_addr=%s\n",
+                            my_snmpTrapAddress_index,
+                            last_snmpTrapAddress_index,
+                            inet_ntoa(my_ip_addr)));
+
+                if (last_snmpTrapAddress_index < my_snmpTrapAddress_index)
+                    last_snmpTrapAddress_index = my_snmpTrapAddress_index;
+
+                /* Detect forwarding loop. */
+                if (var->val_len < 4) {
+                    snmp_log(LOG_ERR, "Length of IP address of OID .1.3.6.1.6.3.18.1.3.%d in PDU is less than %d bytes = %d\n",
+                             my_snmpTrapAddress_index, 4,
+                             (int)var->val_len);
+                } else {
+                    if (to2 &&
+                        memcmp(var->val.string, &my_ip_addr, 4) == 0) {
+                        snmp_log(LOG_ERR, "Forwarding loop detected, OID .1.3.6.1.6.3.18.1.3.%d already has this forwarder's IP address=%s, not forwarding this trap\n",
+                                 my_snmpTrapAddress_index,
+                                 inet_ntoa(my_ip_addr));
+                        return 0;
+                    }
+                    if (memcmp(var->val.string, &agent_addr, 4) == 0) {
+                        snmp_log(LOG_ERR, "Forwarding loop detected, OID .1.3.6.1.6.3.18.1.3.%d already has the sender's IP address=%s, not forwarding this trap\n",
+                                 my_snmpTrapAddress_index,
+                                 inet_ntoa(agent_addr));
+                        return 0;
+                    }
+                }
+            }
+        } /* for var in vblist */
+
+        DEBUGMSGTL(("snmptrapd",
+                    "  last_snmpTrapAddress_index=%d, adding index=%d\n",
+                    last_snmpTrapAddress_index, last_snmpTrapAddress_index+1));
+        /* Change the last index of this OID to the next avaiable number. */
+        forwarder_oid[forwarder_oid_len - 1] = last_snmpTrapAddress_index + 1;
+
+        /*
+         * Add forwarder IP address as OID to trap payload. Use the value
+         * from the transport, so if a v1 PDU is sent, the same IP is not
+         * duplicated you want every forwarder to add this OID with its
+         * own IP address.
+         */
+        snmp_pdu_add_variable(pdu2, forwarder_oid, forwarder_oid_len,
+                              ASN_IPADDRESS, (u_char *)&agent_addr, 4);
+    }
+    return 1;
+}
+
 /*
  *  Trap handler for forwarding to another destination
  */
@@ -1009,14 +1030,25 @@ int   forward_handler( netsnmp_pdu           *pdu,
        INFORMS which may require engineID probing */
 
     pdu2 = snmp_clone_pdu(pdu);
+
+    if (netsnmp_ds_get_boolean(NETSNMP_DS_LIBRARY_ID,
+                               NETSNMP_DS_LIB_ADD_FORWARDER_INFO) &&
+        !add_forwarder_info(pdu, pdu2)) {
+        snmp_close(ss);
+        return NETSNMPTRAPD_HANDLER_FAIL;
+    }
+
     if (pdu2->transport_data) {
         free(pdu2->transport_data);
         pdu2->transport_data        = NULL;
         pdu2->transport_data_length = 0;
     }
-    if (!snmp_send( ss, pdu2 )) {
-	snmp_sess_perror("Forward failed", ss);
-	snmp_free_pdu(pdu2);
+
+    ss->s_snmp_errno = SNMPERR_SUCCESS;
+    if (!snmp_send( ss, pdu2 ) &&
+            ss->s_snmp_errno != SNMPERR_SUCCESS) {
+        snmp_sess_perror("Forward failed", ss);
+        snmp_free_pdu(pdu2);
     }
     snmp_close( ss );
     return NETSNMPTRAPD_HANDLER_OK;
@@ -1056,11 +1088,7 @@ snmp_input(int op, netsnmp_session *session,
     netsnmp_variable_list *vars;
     netsnmp_trapd_handler *traph;
     netsnmp_transport *transport = (netsnmp_transport *) magic;
-    int ret;
-    extern netsnmp_trapd_handler *netsnmp_auth_global_traphandlers;
-    extern netsnmp_trapd_handler *netsnmp_pre_global_traphandlers;
-    extern netsnmp_trapd_handler *netsnmp_post_global_traphandlers;
-
+    int ret, idx;
 
     switch (op) {
     case NETSNMP_CALLBACK_OP_RECEIVED_MESSAGE:
@@ -1084,6 +1112,12 @@ snmp_input(int op, netsnmp_session *session,
 	     */
             if (pdu->trap_type == SNMP_TRAP_ENTERPRISESPECIFIC) {
                 trapOidLen = pdu->enterprise_length;
+                /*
+                 * Drop packets that would trigger an out-of-bounds trapOid[]
+                 * access.
+                 */
+                if (trapOidLen < 1 || trapOidLen > OID_LENGTH(trapOid) - 2)
+                    return 1;
                 memcpy(trapOid, pdu->enterprise, sizeof(oid) * trapOidLen);
                 if (trapOid[trapOidLen - 1] != 0) {
                     trapOid[trapOidLen++] = 0;
@@ -1092,6 +1126,9 @@ snmp_input(int op, netsnmp_session *session,
             } else {
                 memcpy(trapOid, stdTrapOidRoot, sizeof(stdTrapOidRoot));
                 trapOidLen = OID_LENGTH(stdTrapOidRoot);  /* 9 */
+                /* Drop packets with an invalid trap type. */
+                if (pdu->trap_type == LONG_MAX)
+                    return 1;
                 trapOid[trapOidLen++] = pdu->trap_type+1;
             }
             break;
@@ -1113,6 +1150,8 @@ snmp_input(int op, netsnmp_session *session,
 		 * Let's look through the full list....
 		 */
 		for ( vars = pdu->variables; vars; vars=vars->next_variable) {
+                    if (vars->type != ASN_OBJECT_ID)
+                        continue;
                     if (!snmp_oid_compare(vars->name, vars->name_length,
                                           snmpTrapOid, OID_LENGTH(snmpTrapOid)))
                         break;
@@ -1125,8 +1164,8 @@ snmp_input(int op, netsnmp_session *session,
 		    return 1;		/* ??? */
 		}
 	    }
-            memcpy(trapOid, vars->val.objid, vars->val_len);
-            trapOidLen = vars->val_len /sizeof(oid);
+            trapOidLen = SNMP_MIN(sizeof(trapOid), vars->val_len) / sizeof(oid);
+            memcpy(trapOid, vars->val.objid, trapOidLen * sizeof(oid));
             break;
 
         default:
@@ -1157,77 +1196,26 @@ t        *     d) any other global handlers
          *  OK - Enough waffling, let's get to work.....
 	 */
 
-        /*
-	 *  a) authentication handlers
-	 */
-        traph = netsnmp_auth_global_traphandlers;
-        DEBUGMSGTL(("snmptrapd", "Running auth trap handlers\n"));
-	while (traph) {
-            if (!netsnmp_trapd_check_auth(traph->authtypes)) {
-                traph = traph->nexth;
-                continue; /* we continue on and skip this one */
-            }
-	    ret = (*(traph->handler))(pdu, transport, traph);
-            if (ret == NETSNMPTRAPD_HANDLER_FINISH)
-                return 1;
-            if (ret == NETSNMPTRAPD_HANDLER_BREAK)
-                break;
-	    traph = traph->nexth;
-	}
+        for( idx = 0; handlers[idx].descr; ++idx ) {
+            DEBUGMSGTL(("snmptrapd", "Running %s handlers\n",
+                        handlers[idx].descr));
+            if (NULL == handlers[idx].handler) /* specific */
+                traph = netsnmp_get_traphandler(trapOid, trapOidLen);
+            else
+                traph = *handlers[idx].handler;
 
-        /*
-	 *  b) pre-specific global handlers
-	 */
-        traph = netsnmp_pre_global_traphandlers;
-        DEBUGMSGTL(("snmptrapd", "Running pre-global trap handlers\n"));
-	while (traph) {
-            if (!netsnmp_trapd_check_auth(traph->authtypes)) {
-                traph = traph->nexth;
-                continue; /* we continue on and skip this one */
-            }
-	    ret = (*(traph->handler))(pdu, transport, traph);
-            if (ret == NETSNMPTRAPD_HANDLER_FINISH)
-                return 1;
-            if (ret == NETSNMPTRAPD_HANDLER_BREAK)
-                break;
-	    traph = traph->nexth;
-	}
+            for( ; traph; traph = traph->nexth) {
+                if (!netsnmp_trapd_check_auth(traph->authtypes))
+                    continue; /* we continue on and skip this one */
 
-        /*
-	 *  c) trap-specific handlers
-	 */
-        DEBUGMSGTL(("snmptrapd", "Running trap specific handlers\n"));
-        traph = netsnmp_get_traphandler(trapOid, trapOidLen);
-	while (traph) {
-            if (!netsnmp_trapd_check_auth(traph->authtypes)) {
-                traph = traph->nexth;
-                continue; /* we continue on and skip this one */
-            }
-	    ret = (*(traph->handler))(pdu, transport, traph);
-            if (ret == NETSNMPTRAPD_HANDLER_FINISH)
-                return 1;
-            if (ret == NETSNMPTRAPD_HANDLER_BREAK)
-                break;
-	    traph = traph->nexth;
-	}
+                ret = (*(traph->handler))(pdu, transport, traph);
+                if(NETSNMPTRAPD_HANDLER_FINISH == ret)
+                    return 1;
+                if (ret == NETSNMPTRAPD_HANDLER_BREAK)
+                    break; /* move on to next type */
+            } /* traph */
+        } /* handlers */
 
-        /*
-	 *  d) other global handlers
-	 */
-        DEBUGMSGTL(("snmptrapd", "Running global handlers\n"));
-        traph = netsnmp_post_global_traphandlers;
-	while (traph) {
-            if (!netsnmp_trapd_check_auth(traph->authtypes)) {
-                traph = traph->nexth;
-                continue; /* we continue on and skip this one */
-            }
-	    ret = (*(traph->handler))(pdu, transport, traph);
-            if (ret == NETSNMPTRAPD_HANDLER_FINISH)
-                return 1;
-            if (ret == NETSNMPTRAPD_HANDLER_BREAK)
-                break;
-	    traph = traph->nexth;
-	}
 
 	if (pdu->command == SNMP_MSG_INFORM) {
 	    netsnmp_pdu *reply = snmp_clone_pdu(pdu);
