@@ -126,31 +126,26 @@
  *  @{
  */
 #include <net-snmp/net-snmp-config.h>
+#include <net-snmp/net-snmp-features.h>
 #include <sys/types.h>
-#if HAVE_STDLIB_H
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
-#if HAVE_NETINET_IN_H
+#ifdef HAVE_NETINET_IN_H
 #include <netinet/in.h>
 #endif
-#if HAVE_STDLIB_H
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
-#if HAVE_STRING_H
+#ifdef HAVE_STRING_H
 #include <string.h>
 #else
 #include <strings.h>
 #endif
-#if HAVE_WINSOCK_H
-#include <winsock.h>
-#endif
 
-#if HAVE_UNISTD_H
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
-#if HAVE_DMALLOC_H
-#include <dmalloc.h>
-#endif
 
 #include <net-snmp/types.h>
 #include <net-snmp/output_api.h>
@@ -160,6 +155,14 @@
 
 #include <net-snmp/library/snmp_api.h>
 
+netsnmp_feature_child_of(default_store_all, libnetsnmp);
+
+netsnmp_feature_child_of(default_store_void, default_store_all);
+
+#ifndef NETSNMP_FEATURE_REMOVE_DEFAULT_STORE_VOID
+#endif /* NETSNMP_FEATURE_REMOVE_DEFAULT_STORE_VOID */
+
+
 static const char * stores [NETSNMP_DS_MAX_IDS] = { "LIB", "APP", "TOK" };
 
 typedef struct netsnmp_ds_read_config_s {
@@ -176,12 +179,9 @@ static netsnmp_ds_read_config *netsnmp_ds_configs = NULL;
 static int   netsnmp_ds_integers[NETSNMP_DS_MAX_IDS][NETSNMP_DS_MAX_SUBIDS];
 static char  netsnmp_ds_booleans[NETSNMP_DS_MAX_IDS][NETSNMP_DS_MAX_SUBIDS/8];
 static char *netsnmp_ds_strings[NETSNMP_DS_MAX_IDS][NETSNMP_DS_MAX_SUBIDS];
+#ifndef NETSNMP_FEATURE_REMOVE_DEFAULT_STORE_VOID
 static void *netsnmp_ds_voids[NETSNMP_DS_MAX_IDS][NETSNMP_DS_MAX_SUBIDS];
-
-/*
- * Prototype definitions 
- */
-void            netsnmp_ds_handle_config(const char *token, char *line);
+#endif /* NETSNMP_FEATURE_REMOVE_DEFAULT_STORE_VOID */
 
 /**
  * Stores "true" or "false" given an int value for value into
@@ -290,11 +290,13 @@ netsnmp_ds_set_string(int storeid, int which, const char *value)
      */
     if (netsnmp_ds_strings[storeid][which] == value)
         return SNMPERR_SUCCESS;
-    
+
+    snmp_res_lock(MT_LIBRARY_ID, MT_LIB_SESSION);
     if (netsnmp_ds_strings[storeid][which] != NULL) {
         free(netsnmp_ds_strings[storeid][which]);
 	netsnmp_ds_strings[storeid][which] = NULL;
     }
+    snmp_res_unlock(MT_LIBRARY_ID, MT_LIB_SESSION);
 
     if (value) {
         netsnmp_ds_strings[storeid][which] = strdup(value);
@@ -316,6 +318,7 @@ netsnmp_ds_get_string(int storeid, int which)
     return netsnmp_ds_strings[storeid][which];
 }
 
+#ifndef NETSNMP_FEATURE_REMOVE_DEFAULT_STORE_VOID
 int
 netsnmp_ds_set_void(int storeid, int which, void *value)
 {
@@ -324,7 +327,7 @@ netsnmp_ds_set_void(int storeid, int which, void *value)
         return SNMPERR_GENERR;
     }
 
-    DEBUGMSGTL(("netsnmp_ds_set_void", "Setting %s:%d = %x\n",
+    DEBUGMSGTL(("netsnmp_ds_set_void", "Setting %s:%d = %p\n",
                 stores[storeid], which, value));
 
     netsnmp_ds_voids[storeid][which] = value;
@@ -342,6 +345,7 @@ netsnmp_ds_get_void(int storeid, int which)
 
     return netsnmp_ds_voids[storeid][which];
 }
+#endif /* NETSNMP_FEATURE_REMOVE_DEFAULT_STORE_VOID */
 
 int
 netsnmp_ds_parse_boolean(char *line)
@@ -351,6 +355,8 @@ netsnmp_ds_parse_boolean(char *line)
     char           *st;
 
     value = strtok_r(line, " \t\n", &st);
+    if (!value)
+        goto invalid;
     if (strcasecmp(value, "yes") == 0 || 
 	strcasecmp(value, "true") == 0) {
         return 1;
@@ -359,12 +365,14 @@ netsnmp_ds_parse_boolean(char *line)
         return 0;
     } else {
         itmp = strtol(value, &endptr, 10);
-        if (*endptr != 0 || itmp < 0 || itmp > 1) {
-            config_perror("Should be yes|no|true|false|0|1");
-            return -1;
-	}
+        if (*endptr != 0 || itmp < 0 || itmp > 1)
+            goto invalid;
         return itmp;
     }
+
+invalid:
+    config_perror("Should be yes|no|true|false|0|1");
+    return -1;
 }
 
 void
@@ -398,13 +406,17 @@ netsnmp_ds_handle_config(const char *token, char *line)
 
         case ASN_INTEGER:
             value = strtok_r(line, " \t\n", &st);
-            itmp = strtol(value, &endptr, 10);
-            if (*endptr != 0) {
-                config_perror("Bad integer value");
-	    } else {
-                netsnmp_ds_set_int(drsp->storeid, drsp->which, itmp);
-	    }
-            DEBUGMSGTL(("netsnmp_ds_handle_config", "int: %d\n", itmp));
+            if (!value) {
+                config_perror("Missing value");
+            } else {
+                itmp = strtol(value, &endptr, 10);
+                if (*endptr != 0) {
+                    config_perror("Bad integer value");
+                } else {
+                    netsnmp_ds_set_int(drsp->storeid, drsp->which, itmp);
+                }
+                DEBUGMSGTL(("netsnmp_ds_handle_config", "int: %d\n", itmp));
+            }
             break;
 
         case ASN_OCTET_STR:
@@ -442,10 +454,14 @@ netsnmp_ds_register_config(u_char type, const char *ftype, const char *token,
 
     if (netsnmp_ds_configs == NULL) {
         netsnmp_ds_configs = SNMP_MALLOC_TYPEDEF(netsnmp_ds_read_config);
+        if (netsnmp_ds_configs == NULL)
+            return SNMPERR_GENERR;
         drsp = netsnmp_ds_configs;
     } else {
         for (drsp = netsnmp_ds_configs; drsp->next != NULL; drsp = drsp->next);
         drsp->next = SNMP_MALLOC_TYPEDEF(netsnmp_ds_read_config);
+        if (drsp->next == NULL)
+            return SNMPERR_GENERR;
         drsp = drsp->next;
     }
 
@@ -488,10 +504,14 @@ netsnmp_ds_register_premib(u_char type, const char *ftype, const char *token,
 
     if (netsnmp_ds_configs == NULL) {
         netsnmp_ds_configs = SNMP_MALLOC_TYPEDEF(netsnmp_ds_read_config);
+        if (netsnmp_ds_configs == NULL)
+            return SNMPERR_GENERR;
         drsp = netsnmp_ds_configs;
     } else {
         for (drsp = netsnmp_ds_configs; drsp->next != NULL; drsp = drsp->next);
         drsp->next = SNMP_MALLOC_TYPEDEF(netsnmp_ds_read_config);
+        if (drsp->next == NULL)
+            return SNMPERR_GENERR;
         drsp = drsp->next;
     }
 
@@ -522,11 +542,12 @@ netsnmp_ds_register_premib(u_char type, const char *ftype, const char *token,
 }
 
 void
-netsnmp_ds_shutdown()
+netsnmp_ds_shutdown(void)
 {
     netsnmp_ds_read_config *drsp;
     int             i, j;
 
+    snmp_res_lock(MT_LIBRARY_ID, MT_LIB_SESSION);
     for (drsp = netsnmp_ds_configs; drsp; drsp = netsnmp_ds_configs) {
         netsnmp_ds_configs = drsp->next;
 
@@ -550,5 +571,6 @@ netsnmp_ds_shutdown()
             }
         }
     }
+    snmp_res_unlock(MT_LIBRARY_ID, MT_LIB_SESSION);
 }
 /**  @} */
